/**
   Simple test case for rendering the Jacobian in texture space for comparison.
   This program is useful to illustrate that the jacobian will be different 
   if the normalization used in the factorization is ignored....

   Neil Birkbeck
 */
#include "jacobian.h"
#include "utils.h"


/**
   This is a test function for computing the derivatives (using multiple methods)
   for the sequence given the current parameters.

   This was used as a preliminary check that the Jacobian is the same.
   Returns an image.
*/
std::vector<Imagef> derivatives(const Mesh & mesh, 
				const seq_t & seq, 
				const nacb::Imagef & texture,
				Matrix & curParms, 
				bool textureSpace,
				bool updateNormalization = true) {
  const double eps = 1e-5;
  Matrix refP = seq.A*Matrix::eye(3, 4);
  Matrix oE = parms_to_matrix(curParms);

  vector<Matrix> Eup(6, Matrix::zeros(4, 4));
  vector<Matrix> Ederiv(6, Matrix::zeros(4, 4));
  vector<Matrix> matup(6, Matrix::zeros(6, 1));
  int nparms = curParms.m*curParms.n;

  int nchannels = std::min(seq.image.nchannels, std::min(3, texture.nchannels));
  int minchan = 0;
  
  vector<Imagef> grad(nchannels, Imagef(1, 1, 1));
  vector<Imagef> Tgrad(nchannels, Imagef(1, 1, 1));

  for (int k=0; k<nchannels; k++) {
    grad[k] = seq.image.gradient(k + minchan);
    Tgrad[k] = texture.gradient(k + minchan);
    
    // FIXME: should really be scaling up the channels.
    Tgrad[k] *= Tgrad[k].w;  
  }
 
  vector<Imagef> tex(nparms);

  for (int pi=0; pi<curParms.m*curParms.n; pi++) {
    Matrix parms = curParms.copy();
    parms[pi] += eps;
    matup[pi] = seq.A*Matrix::eye(3, 4)*seq.E*parms_to_matrix(parms);
    Eup[pi] = seq.E*parms_to_matrix(parms);

    tex[pi] = Imagef(texture.w, texture.h, nchannels);
    tex[pi] = 0;

    Ederiv[pi] = (Eup[pi] - seq.E*oE)*(1.0/eps);
  }

  bool simpleTexDeriv = false;

  Matrix Pcomp = seq.A*Matrix::eye(3, 4)*seq.E*oE;

  // Current estimate of rotation parameters.
  Matrix Et = seq.E*oE;
  Matrix Rt = Et.submatrix(0, 0, 3, 3);
  Matrix t = Et.submatrix(0, 3, 3, 1);
  
  Matrix M = buildM(seq, Rt, t, Ederiv);

  for (int ti=0; ti<mesh.tris.size(); ti++) {
    Vec3f e1 = mesh.vert[mesh.tris[ti].vi[1]] - mesh.vert[mesh.tris[ti].vi[0]];
    Vec3f e2 = mesh.vert[mesh.tris[ti].vi[2]] - mesh.vert[mesh.tris[ti].vi[0]];
    Vec3f c = mesh.vert[mesh.tris[ti].vi[0]];
    Vec3f n = e1.cross(e2);
    n.normalize();

    double d = c.dot(n);
    Matrix nd = (Matrix)(n*(1.0/d));
    
    Matrix H0 = linear_homography(refP, mesh, ti);
    Matrix H1 = linear_homography(seq.A, Et, c, n);
    
    Matrix Tinv = (H1*H0).inverse();
        
    // Derivatives of the F matrix.
    vector<Matrix> Fderiv(nparms, Matrix::zeros(6, 1));
    for (int pi=0; pi<nparms; pi++) {
      Matrix Up = linear_homography(seq.A, Eup[pi], c, n);
      Fderiv[pi] = ((Up*H0 - H1*H0)*(1.0/eps));
      //Fderiv[pi].printMatlab("Fd");
    }
    //H1.printMatlab("H1");

        
    for (double a=0.02; a<=0.98; a+=1.0/192) {
      for (double b=0.02; b<=0.98-a; b+=1.0/192) {
	double c = 1.0 - a - b;
	Vec3f point = mesh.vert[mesh.tris[ti].vi[0]]*a
	  + mesh.vert[mesh.tris[ti].vi[1]]*b
	  + mesh.vert[mesh.tris[ti].vi[2]]*c;
	Vec3f tc = mesh.tvert[mesh.tris[ti].tci[0]]*a
	  + mesh.tvert[mesh.tris[ti].tci[1]]*b
	  + mesh.tvert[mesh.tris[ti].tci[2]]*c;
	
	Vec3f torig = tc;
	Vec3f tcn = tc;
	
	tc.x *= texture.w;
	tc.y *= texture.h;
	
	
	if (textureSpace && !simpleTexDeriv) {
	  PointCache cache(mesh, ti, Vec3f(a, b, c), H0, seq, texture, Tgrad);
	  if (updateNormalization)
	    cache.updateNormalization(Rt, t);

	  Matrix deriv = (cache.S * M) * (1.0 / cache.normalization);

	  for (int k=0; k<6; k++) {
	    for (int h=0; h<nchannels; h++) {
	      tex[k]((int)tc.x, (int)tc.y, h) = deriv(h, k);
	    }
	  }
	} 
	else {

	  for (int k=0; k<nparms; k++) {
	    float res[nchannels];
	    
	    if (!textureSpace) {	    
	      float zout = 0;
	      Vec2f proj = project(Pcomp, point, &zout);
	      Vec2f pup = project(matup[k], point);
	      
	      Vec2f deriv = (pup - proj)*(1.0/eps);
	      Vec2f imd;
	      
	      for (int h=0; h<nchannels; h++) {
		grad[h].bilinear(proj.x, proj.y, imd.data);
		
		res[h] = imd.dot(deriv);	    
	      }
	    }
	    else if (simpleTexDeriv) {
	      Matrix Jt;
	      Matrix Pd = pd_matrix(torig); 
	      
	      Matrix F = (Fderiv[k]*((Matrix)torig));	    
	      Jt = (Pd * Tinv * F);	    
	      
	      for (int h=0; h<nchannels; h++) {
		Vec2f Timd;
		Tgrad[h].bilinear(round(tc.x), round(tc.y), Timd.data);
		
		res[h] = Timd.dot(Jt);
	      }
	    }	  
	    for (int h=0; h<nchannels; h++) {
	      tex[k]((int)tc.x, (int)tc.y, h) = res[h];
	    }
	  }
	}
      }
    }
  }
  return tex;
}


void test_derivatives(const char * seqname, const char * obj, const char * im, bool updateNormalization) {
  seq_t seq(seqname);
  Mesh mesh;
  Imagef image;

  mesh.loadobj(obj);
  image.load(im);

  image = image.boxFilter(1);
  image = image.flip();

  Matrix curParms = Matrix::zeros(6, 1);

  int nimages = 3;

  for (int posi=0; posi<4; posi++) {
    printf("on pos: %d\n", posi);
    
    Matrix t = (Matrix::random(3, 1) - Matrix::ones(3, 1)*0.5)*0.1;
    Matrix r = (Matrix::random(3, 1) - Matrix::ones(3, 1)*0.5)*0.1;
    Matrix curMat = Matrix::trans(t[0], t[1], t[2])*Matrix::rotz(r[0])*Matrix::roty(r[1])*Matrix::rotx(r[2]);
    curParms = extract_parms(curMat);
    
    for (int imagei=0; imagei<nimages; imagei++) {
      seq.load(imagei);
      seq.E = seq.E*curMat.inverse();
      seq.P = seq.A*Matrix::eye(3, 4)*seq.E;

      vector<Imagef> imageBased = derivatives(mesh, seq, image, curParms, false, updateNormalization);
      vector<Imagef> textureBased = derivatives(mesh, seq, image, curParms, true, updateNormalization);

      for (int k=0; k<textureBased.size(); k++) {
	double n, x;
	textureBased[k].getRange(n, x);

	char textureName[1024], imageName[1024];
	snprintf(textureName, 1024, "/tmp/pos-%d-img-%d-%d-t.png", posi, imagei, k);
	snprintf(imageName, 1024, "/tmp/pos-%d-img-%d-%d-i.png", posi, imagei, k);

	((textureBased[k] - n)*(1.0/(x - n))).write(textureName);
	((imageBased[k] - n)*(1.0/(x - n))).write(imageName);
      }
    } 
  }
}


int main(int ac, char * av[]){
  if (ac <= 3) {
    printf("Need the sequence as an argument, mesh geometry, and a png file.\n");
    return 0;
  }
  
  bool updateNormalization = true;
  if (ac > 4)
    updateNormalization = atoi(av[4]);
   
  test_derivatives(av[1], av[2], av[3], updateNormalization);
  
  return 0;
}
