#include "tracker.h"

Tracker::Tracker(std::vector<std::string> seqNames, 
		 const char * meshName, const char * texName) {
  imagei = 0;
  nchannels = 3;
  
  for (int i=0; i<seqNames.size(); i++) {
    seqs.push_back(seq_t(seqNames[i]));
  }
  loadImages();
  
  mesh.loadobj(meshName);
  
  texture = Imagef(texName);
  texture = texture.flip();
  texture = texture.boxFilter(4);
  
  Tgrad = std::vector<Imagef>(nchannels, Imagef(1, 1, 1));
  
  // Scale derivatives, should do channels individually
  for (int k=0; k<nchannels; k++) {
    Tgrad[k] = texture.gradient(k);
    Tgrad[k] *= Tgrad[k].w;  
  }
  
  parms = extract_parms(Matrix::eye(4, 4));
  points = extract_sample_points(mesh, texture, Tgrad, seqs);

  fbo = new FrameBufferObject(1024, 1024);
  
  bounds = BoundingBox(mesh.vert);
  
  drawProjectedMesh = true;
  drawImageTextures = true;
  
  glGenTextures(1, &texId);
  glBindTexture(GL_TEXTURE_2D, texId);
  texture.initTexture();
  
  for (int i=0; i<seqs.size(); i++) {
    GLuint tex;
    glGenTextures(1, &tex);
    imageTextures.push_back(tex);
  }  
}


Tracker::~Tracker() {
  delete fbo;
  glDeleteTextures(1, &texId);
}


std::vector<std::vector<std::pair<int, int> > > Tracker::updateVisibility() {
  TimeChart::instance().begin("vis");
  
  fbo->bind(1);
  
  nacb::StopWatch timer;
  glClearDepth(1.0);
    
  vector<set<int> > visible_last(seqs.size());
  for (int si=0; si<visible.size(); si++) {
    for (int i=0; i<visible[si].size(); i++) {
      visible_last[si].insert(visible[si][i]);
    }
  }
  visible.clear();

  std::vector<std::vector<std::pair<int, int> > > changed_visibility(seqs.size());

  for (int si=0; si<seqs.size(); si++) {
    vector<int> visi;
    seq_t & seq = seqs[si];
    Matrix oE = parms_to_matrix(parms);
    Matrix P = seq.P*oE;
    Vec2f range = bounds.getBoundsForMatrix(seq.P * oE);
    Vec3f cc = ((seq.E * oE).inverse()).submatrix(0, 3, 3, 1);

    glViewport(0, 0, seq.image.w, seq.image.h);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    cameraPerspectiveMatrix(seq.A(0, 0), seq.A(1, 1),
			    seq.A(0, 2), seq.A(1, 2),
			    seq.image.w, seq.image.h,
			    range.x, range.y, 1);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
      
    Matrix E = (seq.E * oE).transpose();
    glMultMatrixd(E.data);

    mesh.draw();

    glFlush();
    nacb::Imagef vis(seq.image.w, seq.image.h, 1);
    vis = 0;
    glReadPixels(0, 0, vis.w, vis.h, GL_DEPTH_COMPONENT, vis.typeToGL(), vis.data);

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    for (int i=0; i<points.size(); i++) {
      float zout = 0;
      Vec2f pp = project(P, points[i].point, &zout);
      zout = (-zout*(range.x+range.y)/(range.x-range.y) + 2.0*range.x*range.y/(range.x-range.y))/zout;
      zout = (zout + 1.0)/2.0;
	
      Vec3f v = cc - points[i].point;
      v.normalize();
	
      float zread = vis.bilinear(pp.x, pp.y);
      bool pointVis = (pp.x >= 5 && pp.y >= 5) && (pp.x < seq.image.w-5) &&
	(pp.y < seq.image.h-5) && (v.dot(points[i].point) > 0.1) && (zout < zread + kVisibilityTolerance);

      if (pointVis && !visible_last[si].count(i)) {
	changed_visibility[si].push_back(make_pair(i, 1));
      }

      if (!pointVis && visible_last[si].count(i)) {
	changed_visibility[si].push_back(make_pair(i, 0));
      }
       
      if (pointVis)
	visi.push_back(i);	
    }
    visible.push_back(visi);
  }

  printf("Took: %lf\n", double(timer));

  fbo->bind(0);
  TimeChart::instance().end("vis");
  return changed_visibility;
}

  
void Tracker::draw() {
  glPushMatrix();
  Matrix m = parms_to_matrix(parms);
  m = m.transpose();
  glMultMatrixd(m.data);

  glColor3f(1, 1, 1);
  glBindTexture(GL_TEXTURE_2D, texId);
  glEnable(GL_TEXTURE_2D);
  mesh.draw();
  glDisable(GL_TEXTURE_2D);

  if (visible.size()) {
    glEnable(GL_BLEND);

    for (int si=0; si<seqs.size(); si++) {
      glColor4f(si==0, si==1, si==2, 0.5f);
      glBegin(GL_POINTS);
      for (int i=0; i<visible[si].size(); i++) {
	int pi = visible[si][i];
	glVertex3fv(points[pi].point.data);
      }
      glEnd();
    }
  }
   
  glPopMatrix();

  glDisable(GL_TEXTURE_2D);
  glColor3f(1, 1, 1);

  for (int i=0; i<seqs.size(); i++) {
    DrawCamera drawCamera(seqs[i].A, seqs[i].E, seqs[i].image.w, seqs[i].image.h);

    glBindTexture(GL_TEXTURE_2D, imageTextures[i]);
    drawCamera.setEnableTexture(1);
    drawCamera.draw();
  }

  if (drawProjectedMesh) {
      
    for (int i=0; i<seqs.size(); i++) {
      glPushMatrix();

      vector<Vec3f> vert = mesh.vert;
      Matrix Einv = (seqs[i].E).inverse();
      glMultMatrixd(Einv.transpose().data);

      Matrix P = seqs[i].A.inverse()*seqs[i].P* parms_to_matrix(parms);
      for (int j=0; j<mesh.vert.size(); j++) {
	Vec2f co = project(P, mesh.vert[j]);
	mesh.vert[j] = Vec3f(co.x, co.y, 1);
      }

      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      mesh.draw();
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

      mesh.vert = vert;

      glPopMatrix();
    }
  }    
}


void Tracker::imageTrack(int its) {
  TimeChart::instance().begin("image");
    
  static int cnt = 0;
  vector<vector<pair<int, int> > > changed_visibility;

  if (image_grad.size() != seqs.size())
    computeImageGradients();

  if (visible.size() != seqs.size() || (cnt % kUpdateVisibilityEvery) == 0) 
    changed_visibility = updateVisibility();

  for (int i=0; i<its; i++) {
    nacb::Timer timer;
    Matrix J(1, 1);
    Matrix diff(1, 1);
      
    for (int si=0; si<seqs.size(); si++) {
      Matrix diffi(1, 1);
	
      Matrix Ji = image_derivative(points, image_grad[si], seqs[si], 
				   parms, nchannels, diffi, &(visible[si]));
      diffi *= -1;

      if (si == 0) {
	J = Ji;
	diff = diffi;
      }
      else {
	J = Matrix::cat(0, J, Ji);
	diff = Matrix::cat(0, diff, diffi);
      }

      printf("vis[%d] = %d, ", si, visible[si].size());
    }
    printf("\n");

    Matrix upd = Matrix::LlinLeastSq(J, diff); 
    parms += upd;
    //upd.printMatlab("upd");
    //printf("Took: %lf\n", double(timer));

    if (upd.dot(upd) <= kUpdateThreshold) {
      printf("Early break.\n");
      break;
    }
      
  }
  cnt ++;

  TimeChart::instance().end("image");
}


void Tracker::textureTrack(int its) {
  TimeChart::instance().begin("texture");

  static int cnt = 0;
  vector<vector<pair<int, int> > > changed_visibility;
  if (visible.size() != seqs.size() || (cnt % kUpdateVisibilityEvery) == 0) {
    printf("Updated visibility.\n");
    changed_visibility = updateVisibility();
  }

  static std::vector<Matrix> Ls;
  static int updateL = 0;
  static int updateLevery = 10;
    
  if (Ls.size() == 0 || (updateL % updateLevery) == 0) {
    Ls.clear();
    for (int i=0; i<seqs.size(); i++)
      Ls.push_back(Matrix::zeros(210, 210));
  }

  for (int i=0; i<its; i++) {
    Matrix oE = parms_to_matrix(parms);

    Matrix A = Matrix::zeros(6, 1);;
    Matrix N = Matrix::zeros(6, 6);
      

    for (int si=0; si<seqs.size(); si++) {
      // Current estimate of rotation parameters.
      seq_t & seq = seqs[si];

      Matrix Et = seq.E*oE;
      Matrix Rt = Et.submatrix(0, 0, 3, 3);
      Matrix t = Et.submatrix(0, 3, 3, 1);


      vector<Matrix> Ederiv(6, Matrix::zeros(4, 4));
      int nparms = parms.m * parms.n;
      const double eps = 1e-5;
	
      for (int pi=0; pi<parms.m*parms.n; pi++) {
	Matrix p = parms.copy();
	p[pi] += eps;
	Ederiv[pi] = (seq.E*parms_to_matrix(p) - seq.E*oE)*(1.0/eps);
      }
	
      Matrix M = buildM(seq, Rt, t, Ederiv);
      Matrix Mtrans = M.transpose();

      Matrix P = seq.A * Matrix::eye(3, 4) * Et;

      for (int pi=0; pi<points.size(); pi++) {
	points[pi].caches[si].updateNormalization(Rt, t);
      }

      bool check = false;
      bool fullj = false || check;

      Matrix J;
      Matrix diff;

      if (fullj) {
	J = Matrix::zeros(3*points.size(), 6);
	diff = Matrix::zeros(3*points.size(), 1);
	  
	for (int pi=0; pi<points.size(); pi++) {
	  J.setSubmatrix(3*pi, 0, (points[pi].caches[si].S * M) * 
			 (1.0/points[pi].caches[si].normalization)) ;
	    
	  Vec3f color;
	  Vec2f proj = project(P, points[pi].point);
	    
	  seq.image.bilinear(proj.x, proj.y, color.data);
	  color -= points[pi].textureColor;
	  diff[3*pi  ] = color.x;
	  diff[3*pi+1] = color.y;
	  diff[3*pi+2] = color.z;	 	  
	}
      }

      if (check) {
	diff *= -1;

	Matrix diffi;
	Matrix Ji = image_derivative(points, image_grad[si], seqs[si], parms, nchannels, diffi);
	std::vector<Imagef> textureBased = jacobian_to_images(J, points);
	std::vector<Imagef> imageBased = jacobian_to_images(Ji, points);
	  
	for (int k=0; k<6; k++) {
	  double n, x;
	  textureBased[k].getRange(n, x);
	    
	  char textureName[1024], imageName[1024];
	  snprintf(textureName, 1024, "/tmp/J-%d-t.png", k);
	  snprintf(imageName, 1024, "/tmp/J-%d-i.png", k);
	    
	  ((textureBased[k] - n)*(1.0/(x - n))).write(textureName);
	  ((imageBased[k] - n)*(1.0/(x - n))).write(imageName);
	}
	  
	Matrix upd = Matrix::LlinLeastSq(J, diff);
	  
	Matrix upd_i = Matrix::LlinLeastSq(Ji, diff);

	upd.printMatlab("upd");
	upd_i.printMatlab("upd_i");
	  
	printf("res: %lf\n", diff.dot(diff));
      }
      else  {
	/* 
	   Updating L is extremely slow.
	*/
	nacb::Timer timer;
	if (updateL % updateLevery == 0)
	  printf("Updating L\n");

	for (int i=0; i<visible[si].size(); i++) {
	  int pi = visible[si][i];

	  points[pi].caches[si].updateA(seq, P, Mtrans);	    
	  points[pi].caches[si].accumulateA(A);	  
	}
	  
	if (updateL % updateLevery == 0) {
	  for (int pi=0; pi<points.size(); pi++)
	    points[pi].caches[si].accumulateL(Ls[si]);
	}
	else if (changed_visibility.size()){
	  // If not updating completely, just accumulate those that changed.

	  for (int pi=0; pi<changed_visibility[si].size(); pi++) {
	    pair<int, int> point = changed_visibility[si][pi];
	      
	    if (point.second > 0) 
	      points[pi].caches[si].accumulateL(Ls[si]);

	    if (point.second < 0) 
	      points[pi].caches[si].deaccumulateL(Ls[si]);
	  }
	}

	printf("Updating took: %lf\n", double(timer));
	N += Mtrans*Ls[si]*M;
      }	
    }
    Matrix upd = N.inverse()*A*-1;
    upd.printMatlab("upd");
      
    parms += upd;

    if (upd.dot(upd) <= kUpdateThreshold) {
      printf("Early break.\n");
      break;
    }      
  }
  cnt++;
  updateL++;

  TimeChart::instance().end("texture");
}


void Tracker::loadImages() {
  TimeChart::instance().begin("load");

  for (int i=0; i<seqs.size(); i++) 
    seqs[i].load(imagei);

  image_grad.clear();    

  if (drawImageTextures) {
    for (int i=0; i<seqs.size(); i++) {
      glBindTexture(GL_TEXTURE_2D, imageTextures[i]);
      seqs[i].image.initTexture();
    }
    glBindTexture(GL_TEXTURE_2D, 0);
  }

  TimeChart::instance().end("load");
}


void Tracker::computeImageGradients() {
  TimeChart::instance().begin("grad");

  image_grad.clear();

  for (int i=0; i<seqs.size(); i++) {
    vector<Imagef> thisGrad;
      
    for (int k=0; k<nchannels; k++) {
      thisGrad.push_back(seqs[i].image.gradient(k));
    }
    image_grad.push_back(thisGrad);
  }

  TimeChart::instance().end("grad");
}


bool Tracker::trackerLoopIteration(bool imageBased){

  if (seqs[0].image.w <= 1)
    return false;

  loadImages();
    
  if (seqs[0].image.w <= 1)
    return false;
    
  if(imageBased)
    imageTrack();
  else
    textureTrack();
  imagei++;

  return true;
}

