#include "utils.h"

/**
   Convert a matrix to rodrigiues + translation in a vector.
*/
Matrix extract_parms(const Matrix & e) {
  return Matrix::cat(0, (e.submatrix(0, 0, 3, 3)).matrix2rod(), 
		     e.submatrix(0, 3, 3, 1));
}


/**
   Convert the rodrigues parameters to a matrix.
*/
Matrix parms_to_matrix(const Matrix & e) {
  Matrix m = Matrix::eye(4, 4);
  m.setSubmatrix(0, 0, e.submatrix(0, 0, 3, 1).rod2matrix());
  m.setSubmatrix(0, 3, e.submatrix(3, 0, 3, 1));
  return m;
}


/**
   Quick function for affine multiplication.
*/
Vec3f affine3D(const Matrix & P, const Vec3f & co) {
  return Vec3f(P(0, 0)*co.x + P(0, 1)*co.y + P(0, 2)*co.z + P(0, 3),
	       P(1, 0)*co.x + P(1, 1)*co.y + P(1, 2)*co.z + P(1, 3),
	       P(2, 0)*co.x + P(2, 1)*co.y + P(2, 2)*co.z + P(2, 3));	       
}


/**
   Extract the linear homography from the uv-triangle to the 3D triangle.
   \param P    the projection matrix of the camera to conver to (can be eye(3, 4)).
   \param mesh the mesh to source the triangle.
   \param t    the triangle index.
*/
Matrix linear_homography(const Matrix & P, const Mesh & mesh, int t){
  Matrix A0(3, 3), A1(3, 3);
  for (int j=0; j<3; j++){
    A0(0, j) = mesh.tvert[mesh.tris[t].tci[j]].data[0];
    A0(1, j) = mesh.tvert[mesh.tris[t].tci[j]].data[1];
    A0(2, j) = mesh.tvert[mesh.tris[t].tci[j]].data[2];

    Vec3f co = affine3D(P, mesh.vert[mesh.tris[t].vi[j]]); 
    A1(0, j) = co.data[0];
    A1(1, j) = co.data[1];
    A1(2, j) = co.data[2];
  }
  
  // A1 = H * A0;
  Matrix H = A1 * A0.inverse();
  return H;
}


/**
   Extract the linear homography for the triangle in camera P0 to camera P1.
   Return homography taking the point in camera P0 to camera P1.

   \param P0 the projection matrix in P0
   \param P1 the projection matrix in P1
   \param mesh the mesh to source from.
   \param t    the triangle index.
*/
Matrix linear_homography(const Matrix & P0, const Matrix & P1, const Mesh & mesh, int t){
  Matrix A0(3, 3), A1(3, 3);

  for (int j=0; j<3; j++){
    Vec3f co0 = affine3D(P0, mesh.vert[mesh.tris[t].vi[j]]); 
    A0(0, j) = co0.x;
    A0(1, j) = co0.y;
    A0(2, j) = co0.z;

    Vec3f co1 = affine3D(P1, mesh.vert[mesh.tris[t].vi[j]]); 
    A1(0, j) = co1.data[0];
    A1(1, j) = co1.data[1];
    A1(2, j) = co1.data[2];
  }

  
  // A1 = H * A0;
  return A1 * A0.inverse();
}


/**
   Extract the homography for the point c with normal n to the camera 
   with projection matrix [K * eye(3, 4) * E]
*/
Matrix linear_homography(const Matrix & K, const Matrix & E, const Vec3f & c, const Vec3f & n){
  Matrix R = E.submatrix(0, 0, 3, 3);
  Matrix t = E.submatrix(0, 3, 3, 1);
  
  double d = c.dot(n);
  Matrix nmat = n*(1.0/d);
  return  K*(R + t * nmat.transpose())*K.inverse();
}


/**
   Return a vectorized version of the matrix.
*/
Matrix vectorize(const Matrix & m) {
  Matrix r = m.transpose();
  r.m = m.m*m.n;
  r.n = 1;
  return r;
}

/**
   Multiply by the point by homography matrix and and project (Euclidean)
*/
Vec2f Hwarp(Matrix & H, Vec3f & p){
  Matrix X = H*Matrix(p);;
  return Vec2f(X[0]/X[2], X[1]/X[2]);
}

/**
   The partial derivative of the Projection operator of the
   coordinates co
*/
Matrix pd_matrix(nacb::Vec3f & co){
  Matrix Pd = Matrix::zeros(2, 3);
  Pd(0, 0) = 1.0/co.z;
  Pd(1, 1) = 1.0/co.z;
  Pd(0, 2) = -co.x/(co.z*co.z);
  Pd(1, 2) = -co.y/(co.z*co.z);
  return Pd;
}


/**
   Kronecker product of a matrix.
 */
Matrix kronecker_product(const Matrix & m1, const Matrix & m2) {
  Matrix R(m1.m*m2.m, m1.n*m2.n);
  for (int ii=0; ii<m1.m; ii++) {
    for (int jj=0; jj<m1.n; jj++) {
      for (int i=0; i<m2.m; i++) {
	for (int j=0; j<m2.n; j++) {
	  R(ii*m2.m + i, jj*m2.n + j) = m1(ii, jj) * m2(i, j);
	}
      }
    }
  }
  return R;
}

