/**
   Some matrix utility functions.

   Neil Birkbeck
 */
#ifndef UTILS_H
#define UTILS_H

#include <nmath/matrix.h>
#include <nmath/vec3.h>
#include <ik/mesh.h>

using namespace nacb;

/**
   Convert a matrix to rodrigiues + translation in a vector.
*/
Matrix extract_parms(const Matrix & e);


/**
   Convert the rodrigues parameters to a matrix.
*/
Matrix parms_to_matrix(const Matrix & e);


/**
   Quick function for affine multiplication.
*/
Vec3f affine3D(const Matrix & P, const Vec3f & co);


/**
   Extract the linear homography from the uv-triangle to the 3D triangle.
   \param P    the projection matrix of the camera to conver to (can be eye(3, 4)).
   \param mesh the mesh to source the triangle.
   \param t    the triangle index.
*/
Matrix linear_homography(const Matrix & P, const Mesh & mesh, int t);


/**
   Extract the linear homography for the triangle in camera P0 to camera P1.
   Return homography taking the point in camera P0 to camera P1.

   \param P0 the projection matrix in P0
   \param P1 the projection matrix in P1
   \param mesh the mesh to source from.
   \param t    the triangle index.
*/
Matrix linear_homography(const Matrix & P0, const Matrix & P1, const Mesh & mesh, int t);


/**
   Extract the homography for the point c with normal n to the camera 
   with projection matrix [K * eye(3, 4) * E]
*/
Matrix linear_homography(const Matrix & K, const Matrix & E, const Vec3f & c, const Vec3f & n);


/**
   Return a vectorized version of the matrix.
*/
Matrix vectorize(const Matrix & m);


/**
   Multiply by the point by homography matrix and and project (Euclidean)
*/
Vec2f Hwarp(Matrix & H, Vec3f & p);


/**
   The partial derivative of the Projection operator of the
   coordinates co
*/
Matrix pd_matrix(nacb::Vec3f & co);


/**
   Kronecker product of a matrix.
 */
Matrix kronecker_product(const Matrix & m1, const Matrix & m2);


#endif // UTILS_H
